/*
** authored by rue
*/


/*
 * Constants / Globals
 */

const ADBLOCK_CONTRACTID = "@mozilla.org/adblock;1";
const ADBLOCK_CID = Components.ID('{34274bf4-1d97-a289-e984-17e546307e4f}');

const CATMAN_CONTRACTID = "@mozilla.org/categorymanager;1";
const JSLOADER_CONTRACTID = "@mozilla.org/moz/jssubscript-loader;1";

const PREFSERVICE_CONTRACTID = "@mozilla.org/preferences-service;1";
const STDURL_CONTRACTID = "@mozilla.org/network/standard-url;1";
const DIRSERVICE_CONTRACTID = '@mozilla.org/file/directory_service;1';

var SubScriptLoader = Components.classes["@mozilla.org/moz/jssubscript-loader;1"].createInstance(Components.interfaces.mozIJSSubScriptLoader);
var console         = Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService);
function log(msg) { console.logStringMessage(msg); }

var GeneratedKey = Math.random()*100000000;
var whiteKey = "_AdblockWhite"+GeneratedKey;

var apiConstants = new Object(); // handle old + new api constants
var blockTypes = [];			// blockable content-policy types
var blockSchemes = [];			// blockable content-policy schemes
var exceptionTypes = [];		// unblocked content-policy types
var exceptionNames = [];		// unblocked node names
var linkTypes = [];				// link-blockable content-policy types
var linkSchemes = [];			// link-blockable href-protocols
var baseTypes = [];				// unalterable content-policy types
var baseNames = [];				// unalterable node names
var frameCounter = 0;			// counter for "unique" obj-tab names

var isPendingRestart  = false;
var isPendingCallback = null;

var isEnabled = false;
var linkCheck = false;			// pref: check parent-links for img / obj's
var pageBlock = false;			// pref: allow page-blocking
var removeAds = false;
var fastCollapse = false;
var frameObjects = false;
var patterns = [];

var windowMediator = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator);
var windowWatcher  = Components.classes["@mozilla.org/embedcomp/window-watcher;1"].getService(Components.interfaces.nsIWindowWatcher);
var lastBrowser = null;
var lastWindow  = null;

var icoRe = /\.ico$/i; // firefox 1.5 seems to pass any hovered-tab's favicon to the current tab; oddly enough
var xpcRe = /XPCNativeWrapper/i;

/*
 * Content policy class definition
 */

var policy = #10={
	// nsIContentPolicy interface implementation
	shouldLoad: function(contentType, contentLocation, requestOrigin, requestingNode, mimeTypeGuess, extra) {

	 	// if it's an unblockable-type, or unknown-scheme, return the normal policy
		if ( !blockTypes[contentType] || !blockSchemes[contentLocation.scheme]
				/*|| ( exceptionTypes[contentType] && exceptionNames[context.nodeName.toLowerCase()] )*/ // applets are back -!
				|| icoRe.test(contentLocation.spec)
			)
			return apiConstants.ACCEPT;

	 	// handle old api
	 	if (apiConstants.oldStyleAPI) { 	// bool is set by adblockInit()
	 		var context = requestOrigin; 	// oldStyleAPI @params: function(contentType, contentLocation, context, wnd)
	 		var wnd = requestingNode; /*requestOrigin=false, requestingNode=false, mimeTypeGuess=false;*/ }
	 	else {
			// "Bad NPObject As Private Data!" -- unfortunately, firefox1.5 passes 'mis-wrapped' object nodes, which kill the js execution thread, irrespective of try-catch, when touching or testing anything about them; even an 'instanceof' test fails. if 1.6 mends this, the only way we'll benefit is to hard-code a browser version test -- nothing can be tested on the object, itself, or backward-compatibility dies.
			// ensure an implicitly-wrapped node
			var makeImplicit = (nativeWrappingProvided && requestingNode && ((contentType==apiConstants.OBJECT)||!(requestingNode instanceof XPCNativeWrapper)) && mimeTypeGuess!=apiConstants.IMMEDIATE ); //!xpcRe.test(requestingNode.constructor.toString())||
			if (makeImplicit) {
				var implicitWrap = {}; wrapToImplicit(requestingNode, implicitWrap, contentType);
				requestingNode=implicitWrap.requestingNode, wnd=implicitWrap.wnd, doc=implicitWrap.doc;
				implicitWrap.requestingNode = implicitWrap.wnd = implicitWrap.doc = null;  }
			
	 		context = requestingNode; 
	 		if (!wnd && requestingNode) {
	 			if (requestingNode instanceof Components.interfaces.nsIDOMWindow) wnd = requestingNode;
	 			else wnd = requestingNode.defaultView || (requestingNode.localName=="browser"?requestingNode.contentWindow:null) || (requestingNode.ownerDocument ? requestingNode.ownerDocument.defaultView : null);  }
			if ( wnd && wnd.content && wnd instanceof Components.interfaces.nsIDOMChromeWindow) wnd = wnd.content; // no ChromeWindows
	 	}
	 	
		// whitelisting + metadata setup
	 	if (!doc && wnd) var doc = wnd.top.document;
	 	if ( doc) {
	 		var Adblock = (doc._Adblock||(doc._Adblock={}));
			if (!doc[whiteKey]) { try{ 
				var url = (contentType==apiConstants.DOCUMENT) ? contentLocation.spec:(wnd.location+"");
				var white = doc[whiteKey] = (isWhite(url)!=null ? true:false);  }catch(e){} }
			else white = doc[whiteKey]; }
		else if (requestOrigin && requestOrigin.spec) white = (isWhite(requestOrigin.spec)!=null ? true:false);
		else log("no doc or req-origin: "+((context&&context.nodeName)?(context+"  ->  "+context.nodeName):'')+'  ->  '+contentLocation.spec);
		
		if (isEnabled && !white) {
			var filterMatch = null;
			
			// if we've already checked this url, use the prior result -- filter-cache
			if (Adblock && Adblock[contentLocation.spec])
				filterMatch = Adblock[contentLocation.spec].filterMatch;
			// if any containing-link matches a filter..
			if ((filterMatch == false || filterMatch == null) && linkCheck && context && wnd && linkTypes[contentType]) filterMatch = checkLinks(contentType, contentLocation, context, wnd);
			// if we didn't match a link, check the element's source
			if (filterMatch == null) filterMatch = isBlocked(contentLocation); // store the matching filter, or null
			
			if (filterMatch && !filterMatch.white) {
				//var node = (mimeTypeGuess==apiConstants.IMMEDIATE) ? context : elementInterface(contentType, context, wnd,mimeTypeGuess);
				var node = elementInterface(contentType, context, wnd);

				// special handling for documents
				if (contentType == apiConstants.DOCUMENT || (contentType != apiConstants.IMAGE && node.nodeName && node.nodeName.toLowerCase()=="#document")) { // ajax-stuff passes no node, through
					// we'll have to comb through the windows+tabs manually, now; looking for this unloaded URI
					if (pageBlock) {
						adblockPageBlock(filterMatch, contentLocation, wnd); 
						return apiConstants.REJECT_REQUEST; } // if page-block is enabled, write a message and stop.
					else return apiConstants.ACCEPT;  } // no need to continue if the page will load (it would only add us to the adblockable-listing)
					
				if (node) {
					// metadata
					storeAdblockData(node, contentLocation, contentType, filterMatch, wnd,doc);
					
					// special handling for applets -- disables by specifying a default class
					if (node.nodeName && node.nodeName.toLowerCase() == "applet") {
						node.setAttribute("AdblockCode", node.getAttribute("code")); // save the original code attribute
						node.setAttribute("code", "java.applet.Applet"); // this is a default class, which does nothing
						if (node.hasAttribute("AdblockFrameName"))
							node.ownerDocument.getElementById(node.getAttribute("AdblockFrameName")).style.visibility = 'hidden'; }// hide tab-frame -- applets
					
					// hide overlay / frame -- on filter-dialog "refiltration" -- not applets, though
					else if (mimeTypeGuess==apiConstants.IMMEDIATE) 
						(context._AdblockOverlay  && (context._AdblockOverlay.style.visibility = 'hidden')),
						(context._AdblockFrame    && (context._AdblockFrame.style.visibility = 'hidden'));
						
					// if it's not a base-element
					if ( ! (baseTypes[contentType] && node.nodeName && baseNames[node.nodeName.toLowerCase()] )) {
						var immediate = apiConstants.oldStyleAPI ? (node._AdblockImmediate) : (mimeTypeGuess == apiConstants.IMMEDIATE); // (extra instanceof Components.interfaces.nsISupports)
						adblockRemoveFast(node, wnd, immediate, contentLocation.spec);  }
				}
				
				var loadNode = removeAds ? apiConstants.REJECT_REQUEST : apiConstants.ACCEPT;
				return loadNode; // bool: "shouldload" node?
			}
		}
		
		// either the node wasn't blocked OR we're not enabled:
		// -- check original policy, and then object-frame addition..
		loadNode = apiConstants.ACCEPT;
		if (loadNode == apiConstants.ACCEPT && contentType != apiConstants.DOCUMENT) { // ..not for Documents, though!
			node = elementInterface(contentType, context, wnd); // 'node' declared prior
			
			if (node) {
				// store filtering-metadata
				storeAdblockData(node, contentLocation, contentType, filterMatch, /*false,*/ wnd,doc); // -> sending 'false', now, rather than 'null'
				// if we're set to, add a frame to all applets + objects + raw-embeds
				node = elementInterface(contentType, context, wnd);
				if (isEnabled && frameObjects 
						&& 'nodeName' in node
						&& (contentType == apiConstants.OBJECT || /embed/.test(node.nodeName.toLowerCase()) ) // objects *and raw-embeds
						&& contentLocation.spec != node.ownerDocument.URL ) // its not a standalone object
					wnd.setTimeout(makeObjectFrame, 0, node,contentLocation.spec,wnd);
			}
		}
		return loadNode;
	},
	
	// this is now for urls that directly load media, and meta-refreshes (before activation)
	shouldProcess: function(contentType, contentLocation, requestOrigin, requestingNode, mimeType, extra) {
	 	/* If + When the time comes... we're ready:
	 	// handle old api
	 	if (apiConstants.oldStyleAPI) { 	// bool is set by adblockInit()
	 		var context = requestOrigin; 	// oldStyleAPI @params: function(contentType, contentLocation, context, wnd)
	 		var wnd = requestingNode; }
	 	else {
	 		wnd = requestingNode.ownerDocument.defaultView;
	 		context = requestingNode; } */
		return apiConstants.ACCEPT;
	},

	// nsISupports interface implementation
	QueryInterface: function(iid) {
		return this;
	}
};

// Factory object
var factory = {
	// nsIFactory interface implementation
	 createInstance: function(outer, iid) {
		if (outer != null) throw Components.results.NS_ERROR_NO_AGGREGATION;
		return policy;
	},

	// nsIObserver + nsIPrefBranchObserver interface implementation
	observe: function(subject, topic, prefName) { 
		// subject: [wrapped nsISupports :: nsIPrefBranch], nsIPrefBranchInternal
		// topic: "changed"
		
		if (topic == "Adblock-SavePrefFile") {
			var prefObj = Components.classes[PREFSERVICE_CONTRACTID].getService(Components.interfaces.nsIPrefService);
			prefObj.savePrefFile(null); // save the prefs to disk
			return; }
		else if (topic == "Adblock-PrefChange" && prefName == "FilterChange") {
			adblockLoadSettings(prefName); // reload -- with prefname
			return; }
		else if (topic == "Adblock-LoadPrefs") {
			adblockLoadSettings("Init"); // initial load
			return; }
		else if (topic == "Adblock-LoadAPIConstants") {
			var appShell = Components.classes["@mozilla.org/appshell/appShellService;1"].getService(Components.interfaces.nsIAppShellService);
			var hiddenWnd = appShell.hiddenDOMWindow; // global hidden-window
			if (!hiddenWnd.apiConstants    ) hiddenWnd.apiConstants     = apiConstants;
			if (!hiddenWnd.GeneratedKey    ) hiddenWnd.GeneratedKey     = GeneratedKey;
			if (!hiddenWnd.isBlocked       ) hiddenWnd.isBlocked        = isBlocked;
			if (!hiddenWnd.isWhite         ) hiddenWnd.isWhite          = isWhite;
			if (!hiddenWnd.addWhite        ) hiddenWnd.addWhite         = addWhite;
			if (!hiddenWnd.storeAdblockData) hiddenWnd.storeAdblockData = storeAdblockData;
			if (!hiddenWnd.exportWhitelist ) hiddenWnd.exportWhitelist  = exportWhitelist;
			if (!hiddenWnd.whiteKey        ) hiddenWnd.whiteKey         = whiteKey;
			hiddenWnd.isPendingRestart  = isPendingRestart;
			hiddenWnd.isPendingCallback = isPendingCallback; 
			return; }
		else if (topic == "Adblock-ImportFilters") { // right-click > selection-import
			importFiltersFromString(prefName);
			return; }
		else {
			var prefRe = /^adblock\.(?!observer|patterns)/; // reloads settings only on appropriate pref-change
			if (prefRe.test(prefName)) adblockLoadSettings(prefName); } // reload -- with prefname
		
	},
	
	// nsISupports interface implementation
	QueryInterface: function(iid) {
		if (!iid.equals(Components.interfaces.nsISupports) &&
			!iid.equals(Components.interfaces.nsISupportsWeakReference) &&
			!iid.equals(Components.interfaces.nsIFactory) &&
			!iid.equals(Components.interfaces.nsIObserver))
		{
			dump("Adblock content policy factory object: QI unknown interface: " + iid + "\n");
			throw Components.results.NS_ERROR_NO_INTERFACE;
		}
		
		return this;
	}
}



/*
 * Core Routines
 */

adblockInit(); // begin initialization

// Initialization and registration
function adblockInit() {
	
	// Standard Pref-observer registration
	try {
		var prefInternal = Components.classes[PREFSERVICE_CONTRACTID].getService(Components.interfaces.nsIPrefBranchInternal);
		prefInternal.addObserver("adblock", factory, true);
	} catch (e) { log("Adblock content policy registration: exception when registering pref observer: " + e + "\n"); }
	
	// Custom Pref-observer registration
	try {
		var observerService = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
		observerService.addObserver(factory, "Adblock-PrefChange", true);
		observerService.addObserver(factory, "Adblock-SavePrefFile", true);
		observerService.addObserver(factory, "Adblock-LoadAPIConstants", true);
		observerService.addObserver(factory, "Adblock-LoadPrefs", true);
		observerService.addObserver(factory, "Adblock-ImportFilters", true);
	} catch (e) { log("Adblock content policy registration: exception when registering saveprefs observer: " + e + "\n"); }

	// Variable initialization
	
	// content-policy api constants : old + new
	apiConstants.oldStyleAPI = typeof(Components.interfaces.nsIContentPolicy.TYPE_DOCUMENT) == 'undefined';
	apiConstants.UNIMPLEMENTED = -72591; // our own, custom value -- hopefully it remains available  ( -used by apiConstants.REFRESH)
	apiConstants.XMLHTTPREQUEST = -72592; // snip -> ajax
	apiConstants.OTHER = 		Components.interfaces.nsIContentPolicy.OTHER 		| Components.interfaces.nsIContentPolicy.TYPE_OTHER;
	apiConstants.SCRIPT = 		Components.interfaces.nsIContentPolicy.SCRIPT 		| Components.interfaces.nsIContentPolicy.TYPE_SCRIPT;
	apiConstants.IMAGE = 		Components.interfaces.nsIContentPolicy.IMAGE 		| Components.interfaces.nsIContentPolicy.TYPE_IMAGE;
	apiConstants.STYLESHEET = 	Components.interfaces.nsIContentPolicy.STYLESHEET 	| Components.interfaces.nsIContentPolicy.TYPE_STYLESHEET;
	apiConstants.OBJECT = 		Components.interfaces.nsIContentPolicy.OBJECT 		| Components.interfaces.nsIContentPolicy.TYPE_OBJECT;
	apiConstants.DOCUMENT = 	Components.interfaces.nsIContentPolicy.DOCUMENT 	| Components.interfaces.nsIContentPolicy.TYPE_DOCUMENT;
	apiConstants.SUBDOCUMENT = 	Components.interfaces.nsIContentPolicy.SUBDOCUMENT 	| Components.interfaces.nsIContentPolicy.TYPE_SUBDOCUMENT;
	apiConstants.REFRESH = 			apiConstants.oldStyleAPI ? apiConstants.UNIMPLEMENTED : Components.interfaces.nsIContentPolicy.TYPE_REFRESH; // new api ONLY :: shouldProcess() handles this
	apiConstants.ACCEPT = 			apiConstants.oldStyleAPI ? true  : Components.interfaces.nsIContentPolicy.ACCEPT; 			// +1
	apiConstants.REJECT_REQUEST = 	apiConstants.oldStyleAPI ? false : Components.interfaces.nsIContentPolicy.REJECT_REQUEST; 	// -1 -- unilateral rejection
	apiConstants.REJECT_TYPE = 		apiConstants.oldStyleAPI ? false : Components.interfaces.nsIContentPolicy.REJECT_TYPE; 		// -2 -- rejected based solely on its type (of the above flags)
	apiConstants.REJECT_SERVER = 	apiConstants.oldStyleAPI ? false : Components.interfaces.nsIContentPolicy.REJECT_SERVER; 	// -3 -- rejected based on hosting/requesting server (aContentLocation or aRequestOrigin)
	apiConstants.REJECT_OTHER = 	apiConstants.oldStyleAPI ? false : Components.interfaces.nsIContentPolicy.REJECT_OTHER; 	// -4 -- maybe direct third-party callers to consult the "extra" param for additional details
	apiConstants.IMMEDIATE = -73; // another custom value (smaller for speed) -- same hope
	apiConstants.NULL = -92; // UNUSED: custom value
	
	// blockable types + schemes
	blockTypes[apiConstants.UNIMPLEMENTED] = false; // DISABLED
	blockTypes[apiConstants.SCRIPT] = true;
	blockTypes[apiConstants.IMAGE] = true;
	blockTypes[apiConstants.OBJECT] = true;
	blockTypes[apiConstants.DOCUMENT] = true;
	blockTypes[apiConstants.SUBDOCUMENT] = true;
	blockTypes[apiConstants.XMLHTTPREQUEST] = true;
	blockSchemes["file"] = false;
	blockSchemes["http"] = true;
	blockSchemes["https"] = true;

	// unblocked types + nodeNames
	exceptionTypes[apiConstants.OBJECT] = false; // DISABLED
	exceptionNames["applet"] = false; // DISABLED

	// link-searchable types + href-protocols
	linkTypes[apiConstants.IMAGE] = true;
	linkTypes[apiConstants.OBJECT] = true;
	linkSchemes["file:"] = false;
	linkSchemes["http:"] = true;
	linkSchemes["https:"] = true;
	linkSchemes["javascript:"] = true;

	// unalterable content-policy types + nodeNames -- root-elements
	baseTypes[apiConstants.IMAGE] = true;
	baseTypes[apiConstants.DOCUMENT] = true;
	baseTypes[apiConstants.SCRIPT] = true;
	baseTypes[apiConstants.XMLHTTPREQUEST] = true;
	baseNames["html"] = true;
	baseNames["body"] = true;
	baseNames["script"] = true;
	baseNames["xmlhttprequest"] = true;

	// Temp-Dir Cleanup (for prior uninstalls) -- separate, in case there's no default profile, or the user isn't root
	var dirArray = new Array();
	dirArray[0] = "AChrom"; 	// app. chrome magic-key
	dirArray[1] = "UChrm"; 		// profile chrome magic-key
	dirArray[2] = "ProfD"; 		// profile directory magic-key
	dirArray[3] = "ComsD"; 		// component directory magic-key
	for (var i = 0, n ; i < dirArray.length ; i++) {
		try { 
			var dirService = Components.classes[DIRSERVICE_CONTRACTID].getService(Components.interfaces.nsIProperties);
			var currentDir = dirService.get(dirArray[i], Components.interfaces.nsIFile);
			var tempDir = currentDir.clone();
			var uninstalledFile = currentDir.clone();
			
			tempDir.append("adblock-temp");
			uninstalledFile.append("adblock-uninstalled");
			n = 1;
			
			// remove temp-dir
			if (tempDir.exists()) tempDir.remove(true);
			
			// remove installer-uninstalled files
			while (uninstalledFile.exists()) {
				uninstalledFile.remove(true);
				uninstalledFile = currentDir.clone();
				uninstalledFile.append("adblock-uninstalled" + n);
				n++; }
		} catch(e) { /*throw(e);*/ }
	}

	// Load Settings
	//adblockLoadSettings("Init"); // XXX NO. if there's no default-profile, this ends up being duplicated, when the first browser-wnd finally loads

}

// lends implicit-wrapping, for:  node + window + document
function wrapToImplicit(requestingNode, implicitWrap, contentType) {
	var requestingNodeUnwrapped = requestingNode;
	var forceWrap = (contentType==apiConstants.OBJECT);
	try { requestingNode = adblockWrap(requestingNode, forceWrap, 'defaultView','top','contentWindow','ownerDocument'); }catch(e) { /*throw(e);*/ }
	
	if (requestingNode instanceof Components.interfaces.nsIDOMWindow) var domWnd = requestingNode;
	else domWnd = requestingNode.defaultView || (requestingNode.localName=="browser"?requestingNode.contentWindow:null) || (requestingNode.ownerDocument ? requestingNode.ownerDocument.defaultView : null);
	
	if (!lastWindow||lastWindow.closed) lastWindow = windowMediator.getMostRecentWindow("navigator:browser");
	
	var newEvent = lastWindow.document.createEvent('MouseEvents'); 
	newEvent.initMouseEvent('abCatchNode', true,true, domWnd, 1, 0,0,0,0, false,false,false,false, 1, requestingNodeUnwrapped);
	implicitWrap.wnd =  newEvent.view;   
	newEvent.initMouseEvent('abCatchNode', true,true, null, 1, 0,0,0,0, false,false,false,false, 1, {});
			newEvent = lastWindow.document.createEvent('MutationEvents'); 
			newEvent.initMutationEvent('abCatchNode2', true,true, implicitWrap.wnd.top.document, "a","b","c",1);
			implicitWrap.doc = newEvent.relatedNode;   // ..we'll skip the re-init/re-create, and jump to next
			//newEvent.initMutationEvent('abCatchNode2', true,true, {}, "a","b","c",1);
					//newEvent = lastWindow.document.createEvent('MutationEvents'); 
					newEvent.initMutationEvent('abCatchNode3', true,true, requestingNodeUnwrapped, "a","b","c",1);
					implicitWrap.requestingNode = newEvent.relatedNode;  
					newEvent.initMutationEvent('abCatchNode3', true,true, {}, "a","b","c",1);
		
	if (!implicitWrap.doc._Adblock) implicitWrap.doc._Adblock={};
}

// returns the queryInterface to a dom-object or frame / iframe -- for 'shouldload' policy-check
function elementInterface(contentType, context, wnd) {
	try {
		//if (contentType == apiConstants.XMLHTTPREQUEST || )
		//	return {tagName:"XMLHttpRequest"};
		//else 
		if (!apiConstants.oldStyleAPI || context && context._AdblockImmediate)
			return context;
		else if (contentType == apiConstants.SUBDOCUMENT)
			return wnd.frameElement;
		else if (contentType == apiConstants.DOCUMENT)
			return wnd.document.documentElement
		else if (contentType == apiConstants.IMAGE && !context)
			return wnd.document; // background-images
		else 
			return context.QueryInterface(Components.interfaces.nsIDOMElement);
	} catch(e) { return context; }
}

var nativeWrappingProvided = typeof(XPCNativeWrapper) == 'function';
function adblockWrap(object,forceWrap) {
	if(!object) throw("Null object passed to Adblock wrap()");
	
	//if(!forceWrap && object.wrappedJSObject || !nativeWrappingProvided) // implicit nativeWrapper already OR no wrapping provided by the browser
		//return object; 
	//else {
		var wrapper = object;
		try { wrapper = new XPCNativeWrapper(object); } // new-style Deep wrapper constructor
		catch(e) { 
			// ..an older build -> use the older Shallow constructor
			try {
				wrapper = new XPCNativeWrapper(object,
						arguments[2], arguments[3], arguments[4],
						arguments[5], arguments[6], arguments[7],
						arguments[8], arguments[9], arguments[10] );
			} 
			catch(e){ log && log(e); } //wrapper={}; }
		}
		return wrapper;
	//}
}

// stores adblock-metadata for an element
function storeAdblockData(node, contentLocation, contentType, filterMatch, wnd,doc) {
	if (!(wnd||doc)) return;
	if (!doc) doc = wnd.top.document; // ..the ajax-intercept uses mutation to get the sometimes-unwrapped .document, so we'll prefer that
	if (!doc) return;
	//if (filterMatch===false)
		//var x = true;
	
	var Adblock = (doc._Adblock||(doc._Adblock={})); // -> switched to doc, instead of wnd, because bfcache kills the window's implicit-wrapper -- but the document's is retained  // (/*wnd._Adblock||*/doc._Adblock||(/*wnd._Adblock=*/doc._Adblock={}))
	
	// for images repeated on page store node for each repeated image
	var data = Adblock[contentLocation.spec];
	if (data) {
		node && data.node.push(node);
		return; //data.node.length-1;
	}
	
	data = {};
	data.node = [];
	node && data.node.push(node);
	data.contentLocation = contentLocation;
	data.contentType = contentType;
	data.filterMatch = filterMatch; //||false; // ..take note: this permits linkCheck without the others, per filter-cache.
	
	Adblock[contentLocation.spec] = data;
	
	//return data.node.length-1;
}

// for blocked / hidden pages 
//   -- writes a message showing the filter and its matched portion of url
function adblockPageBlock(filterMatch, contentLocation, wnd) {
	filterMatch = filterMatch.toString(); // force-convert filter to string
	var urlRe = new RegExp(filterMatch.replace(/^\/(.*)\/i?$/, "($1)"), "i"); // must come before filterMatch's change
	var urlHilight = contentLocation.spec.replace(urlRe, "<span class='urlHilight'>$1</span>");
	var simpleRe = /\(\?\:\\n\)\?/; // test for the "Adblock simple-filter" flag
	if (simpleRe.test(filterMatch)) filterMatch = filterMatch.replace(/^(\@\@)?\/(.*)\(\?\:\\n\)\?\/i?$/, "$1$2").replace(/\\\./g, " ").replace(/\.\*/g, "*").replace(/\s/g, ".").replace(/\\/g, ""); // the first 'replace' is same as substr, below -- left regexp for fun
	else filterMatch = filterMatch.substr(0, filterMatch.length - 1); //.replace(/\\/g, "\\\\"); // length isn't zero-based + / + $ + flag "i" are anchored to end
	var title = contentLocation.host.replace(/^www\./, "");
	//var pageElementLocal = wnd.document.documentElement;
	// write the block-message into the page -- [Note: doesn't center properly for pages with prior content]
	//pageElementLocal.innerHTML = 
	//			<title>Adblocked: "+title+"</title>
	var msgHTML = "\
		<head>\
			<style type='text/css'>\
			<!--\
				.massive 	{ font-family: Times, Times New Roman, Serif; font-size: 400%; color: #999999;}\
				.infotext	{ font-family: Arial, Helvetica, San-serif; font-size: medium; color: #999999;}\
				.filter		{ font-family: Arial, Helvetica, San-serif; font-size: medium; color: #993333;}\
				.url		{ font-family: Arial, Helvetica, San-serif; font-size: small; color: #999999;}\
				.urlHilight	{ border-width: 0 0 1px 0; border-color: default; border-style: none none solid none;}\
			-->\
			</style>\
		</head>\
		<body>\
			<table style='width: 100%; height: 100%;'><tr valign='middle'><td><center>\
				<span class='massive'>Adblock</span><br>\
				<span class='infotext'>This site has been blocked: </span><span class='filter'>"+filterMatch+"</span><br>\
				<span class='url'>"+urlHilight+"</span>\
			</center></td></tr></table>\
		</body>";
	//try {wnd.location = "about:blank";} catch(e) {}
	//try {wnd.location = "chrome://adblock/content/lockup.html";} catch(e) {}
	//wnd.setTimeout("document.documentElement.innerHTML = '"+msgHTML+"';alert('done:\n'+document.nodeName);", 100);
	wnd.setTimeout(pageCallback, 50, wnd,msgHTML);
	// XXX ..here, queryInterface directly to webNav, and adjust the history
	return;
}
function pageCallback(wnd,msgHTML) { wnd.stop(); wnd.document.documentElement.innerHTML = msgHTML; };

// hides a BLOCKED element -- faster / less thorough
function adblockRemoveFast(node, wnd, immediate, path) {
	// serialize iframes - to avoid graphics-glitch -- old api only
	if (apiConstants.oldStyleAPI && node.nodeName.toLowerCase() == "iframe")
		wnd.document.documentElement.innerHTML = "";

	// kill background images
	if (node.toString().indexOf("HTMLDocument")!=-1) {
		path = path.slice(path.lastIndexOf("/")+1, path.length);
		var Sheets = node.styleSheets;
		for (var s in Sheets)
			for (var r in Sheets[s].cssRules) {
				var style = Sheets[s].cssRules[r].style;
				if (style && style.backgroundImage.indexOf(path) != -1) 
					wnd.setTimeout(fastCallback, 0, style);  }
		return;  }
	
	// collapse frameset-cols+rows for frames
	if (node.nodeName.toLowerCase() == "frame" && node.parentNode.nodeName.toLowerCase() == "frameset") {
		var i = 0, prevFrame = node;
		while (prevFrame.previousSibling) {
			if (/^frame|frameset$/i.test(prevFrame.previousSibling.nodeName)) i++;
			prevFrame = prevFrame.previousSibling; }

		/* This was a frameset alteration-listener. it proved too complicated to be desirable
		if (node.parentNode.hasAttribute("adblockrows") || node.parentNode.hasAttribute("adblockcols"))
			return;
			node.parentNode.removeEventListener("DOMAttrModified", onFramesetChange, true); */
		var setTypes = ["cols", "rows"], RowsOrCols;
		for (var t in setTypes) {
			RowsOrCols = setTypes[t];
			if (node.parentNode.hasAttribute(RowsOrCols))
				adblockRemoveSlow(node, wnd, immediate, '\
					var setWidths = node.parentNode.getAttribute("'+RowsOrCols+'").split(",");\
					setWidths['+i+'] = "0";\
					setWidths = setWidths.join(",");\
					node.parentNode.setAttribute("adblock'+RowsOrCols+'", setWidths);\
					node.parentNode.setAttribute("'+RowsOrCols+'", setWidths);\
				'); }
		/*node.parentNode.addEventListener("DOMAttrModified", onFramesetChange, true);*/
	}
	// explicitly hide nodes (when adding a filter)
	else (fastCollapse) ?
		adblockHide(node, wnd, immediate)
		:
		adblockRemoveSlow(node, wnd, immediate);
}
function fastCallback(style){ style.backgroundImage = ""; }


/* We're stuck with code-strings below -- newer builds sandbox anonymous-functions from accessing 'node' */

// hides a BLOCKED element -- slower / more thorough
function adblockRemoveSlow(node, wnd, immediate, removalCode) {
	// use timeout to collapse node - otherwise the reflow-queue doesn't catch it
	// set style + collapse frameset-cols+rows for frames (if specified)
	
	// XXX disabled for bug 274425 -- crashed, when set directly, so we'll always prefer timeout
	//if (immediate) (removalCode) ? eval(removalCode) : (node.style.display = "none");
	//else {
		wnd.setTimeout(slowCallback, 0, node,removalCode);
	//}
}
function slowCallback(node,removalCode){ try{if (removalCode) eval(removalCode); else node.style.display='none';}catch(e){throw node.nodeName||node.localName} }

// hides a LOADED element
function adblockHide(node, wnd, immediate) {
	// use timeout to hide node - so reflow-queue gets it
	try {node = node.QueryInterface(Components.interfaces.nsIDOMElement);} catch(e) {}
	//if (immediate) node.style.setProperty("visibility", "hidden", "important");
	//else {
		wnd.setTimeout(hideCallback, 0, node);
	//}
}
function hideCallback(node){ node.style.setProperty("visibility", "hidden", "important"); }



/*
 * URL checking
 */

// Tests if a given url is whitelisted for either the host, or full-url
var whitelist={},  whitehosts={},  whiteImported={}, stdURLClass=Components.classes[STDURL_CONTRACTID],  nsIURI=Components.interfaces.nsIURI;
function isWhite(url) {
	var stdURL = stdURLClass.createInstance(nsIURI);
		stdURL.spec = url;
	var host = stdURL.host;
		if (host) host = host.replace(/^www\./,"");
	if (whitelist[host])  { whitelist[host] = new Date().valueOf(); return 1; }
	if (whitelist[url]) { whitelist[url] = new Date().valueOf(); return 0; }
	return null;
}

// Tests if any containing-link matches a filter
function checkLinks(contentType, contentLocation, context, wnd) {
	var parentLink = null, parentLinkLocation, parentNode = elementInterface(contentType, context, wnd).parentNode;
	while (parentNode && parentLink == null) {
		if (typeof(parentNode.href) != 'undefined' && parentNode.href 
				&& linkSchemes[parentNode.protocol.toLowerCase()] ) 
			parentLink = parentNode.href;
		//else if (parentNode.getAttribute("URL")) parentLink = parentNode.getAttribute("URL"); // unneeded ?
		else parentNode = parentNode.parentNode; }
	if (parentLink) { try {
		parentLinkLocation = new Object();
		parentLinkLocation.spec = parentLink;
		parentLinkLocation.host = "all"; //}
		var linkMatch = isBlocked(parentLinkLocation); 
		//if (linkMatch) parentNode.removeAttribute("href"); // DISABLED: stops other children from being caught --eliminate link, if matching
		return linkMatch; } catch(e) { wnd.alert("Adblock LinkCheck Error: " + contentLocation.spec + "\n\n" + parentLink + "\n\n" + e); return null;} }
	else 
		return null;
}

// Tests if a given URL should be blocked
function isBlocked(url) {
	// white filters
	if (patterns.white) {
		if (patterns.white[url.host]) {
			var match = checkURL(url.spec, patterns.white[url.host]);
			if (match || url.host == 'all') 
				return (match) ? {white:match,asWhiteString:'@@'+match} : match; /*false;*/ } // if no matches (and not linkCheck), continue..
		if (patterns.white[''])
			if ((match = checkURL(url.spec, patterns.white[''])))
				return {white:match,asWhiteString:'@@'+match}; /*false;*/ // if no matches, check blaclist.
	}
	// black filters
	if (patterns[url.host]) {
		match = checkURL(url.spec, patterns[url.host]);
		if (match || url.host == 'all') 
			return match; } // if no matches (and not linkCheck), continue..
	if (patterns[''])
		return checkURL(url.spec, patterns['']); // if no matches, we're done.
	return null;
}

// Checks if the URL matches any of the patterns from the list
function checkURL(url, patterns) {
	for (var i=0; i<patterns.length; i++)
		if (patterns[i].test(url))
			return patterns[i];

	return null; // if no matches, return null
}



/*
 * Filter + Settings management
 */

// Loads the preferences
function adblockLoadSettings(prefName) {
	var prefObj = Components.classes[PREFSERVICE_CONTRACTID].getService(Components.interfaces.nsIPrefService);
	var Branch = prefObj.getBranch("adblock."), type="getPrefType";
	var boolType = Components.interfaces.nsIPrefBranch.PREF_BOOL, set="setBoolPref";
	var charType = Components.interfaces.nsIPrefBranch.PREF_STRING;
	
	// Set prefs if under Mozilla
	checkPrefs();
	
	// Load Bool-prefs -- on init or inidividual-change
	if (prefName && prefName != "FilterChange") {
		isEnabled    = !(Branch[type]("enabled"     )==boolType) ||  Branch.getBoolPref("enabled");      // default:true
		linkCheck    =  (Branch[type]("linkcheck"   )==boolType) &&  Branch.getBoolPref("linkcheck");    // default:false
		pageBlock    =  (Branch[type]("pageblock"   )==boolType) &&  Branch.getBoolPref("pageblock");    // default:false
		removeAds    = !(Branch[type]("hide"        )==boolType) || !Branch.getBoolPref("hide");         // rev-hide -- default:true
		fastCollapse =  (Branch[type]("fastcollapse")==boolType) &&  Branch.getBoolPref("fastcollapse"); // default:false
		frameObjects = !(Branch[type]("frameobjects")==boolType) ||  Branch.getBoolPref("frameobjects"); // default:true
		
		try{ 
			if (typeof AjaxIntercept == 'undefined') //if (prefName == "Init") 
				SubScriptLoader.loadSubScript("chrome://adblock/content/ajax-intercept.js"); // snip !
			if (isEnabled) AjaxIntercept.startIfEnabled();
			else AjaxIntercept.stop();
		}catch(e){}
		
		// don't continue, if we're not initializing
		if (prefName != "Init") return;
	}
	
	// Load Filters + Whitelist -- on init or list-change
	loadFilters(prefName);
	loadWhitelist(prefName);
}

function loadWhitelist(prefName) {
	var prefObj = Components.classes[PREFSERVICE_CONTRACTID].getService(Components.interfaces.nsIPrefService);
	var Branch = prefObj.getBranch("adblock."), type="getPrefType";
	var charType = Components.interfaces.nsIPrefBranch.PREF_STRING;
	var curTime = new Date().valueOf();

	whitelist = {},  whitehosts = {};
	var list = (Branch[type]("white")==charType) ? Branch.getCharPref("white") : "";
	list = (list!="") ? list.split("   ") : [];
	for (var n in whiteImported) list.unshift([n,curTime].join(" ")); // ..anything our loadFilters() might have imported
	if (!list.length) return;

	var oneSec=1000, oneMin=oneSec*60, oneHour=oneMin*60, oneDay=oneHour*24, oneMonth=oneDay*30, oneYear=oneDay*365, halfYear=oneYear/2;
	var stdURL = Components.classes[STDURL_CONTRACTID].createInstance(Components.interfaces.nsIURI);
	var i=-1, l, url, host, date;
	while ((l=list[++i])) {
		l=l.split(" "), url=l[0], date=l[1], stdURL.spec=url, host=stdURL.host;
			if (host) host = host.replace(/^www\./,"");
		whitehosts[host]||(whitehosts[host]={});  whitehosts[host][url] = true;
		var elapsed = (curTime - parseInt(date));
		var halfYearCount = (elapsed/halfYear);
		var prune = (halfYearCount>1); // auto-prune the list, for entries older than 6months
		if (!prune) whitelist[url] = date; //if (/^http\:\/\//.test(l))
		else {
			delete whitelist[url]; // ..in case an imported entry was already there (unshifted with a new date, above, remember?)
			var wasPruned = true; }
	}
	
	//if (wasPruned) 
	saveWhitelist(); // XXX always save, since we may have imported some whitelist stuff
}
function saveWhitelist() {
	var prefObj = Components.classes[PREFSERVICE_CONTRACTID].getService(Components.interfaces.nsIPrefService);
	var Branch = prefObj.getBranch("adblock.");

	var asArray = exportWhitelist();
	if (asArray.length>0) Branch.setCharPref("white", asArray.join("   "));
	else Branch.setCharPref("white",""), Branch.deleteBranch("white"); // first set it empty, so pref-observers trip (eg. status whitelist-styling update)

	prefObj.savePrefFile(null); // save the prefs to disk
}
function exportWhitelist(expType) {
	switch(expType) {
		case 'asArray':
			var asArray = [], items;
			for (var n in whitelist) asArray.push(n);
			return asArray;
		case 'asFullPrefArray':
		default: 
			asArray = [], items;
			for (n in whitelist) items=[n,whitelist[n]], asArray.push(items.join(" "));
			return asArray;
	}
}

function loadFilters(prefName, listToImport, shouldAppend) {
	var prefObj = Components.classes[PREFSERVICE_CONTRACTID].getService(Components.interfaces.nsIPrefService);
	var Branch = prefObj.getBranch("adblock."), type="getPrefType";
	var charType = Components.interfaces.nsIPrefBranch.PREF_STRING;
	
	patterns = [];
	if (!listToImport || shouldAppend) {
		var list = (Branch[type]("patterns")==charType) ? Branch.getCharPref("patterns") : "";
		list = (list!="") ? list.split(" ") : [];
		if (listToImport && shouldAppend) list = list.concat(listToImport); }
	else list = listToImport;
	
	/*
	// Default Filters (for unset pref)
	if (!list.length) {
		SubScriptLoader.loadSubScript("chrome://adblock/content/component-defaultfilters.js"); // gives us "var defaultFilterString"
		Branch.setCharPref('patterns', defaultFilterString);
		list = defaultFilterString; }
	*/
	
	if (!list.length) return;

	// remove duplicate entries
	var url = Components.classes[STDURL_CONTRACTID].createInstance(Components.interfaces.nsIURI);
	var o = {}, i=-1, l;
	while ((l=list[++i])) { if (o[l]) delete o[l]; o[l]=true; } // this retains the recent-most occurrence of any duplicates
	list = [];
	for (var n in o) list.push(n);
	
	// load list into memory
	whiteImported = {};
	var whiteRe = /^\#\#/, patRe = /^\/.*\/$/, simpRe = /\*/, periodRe = /\./;
	var wpRe = /^\@\@/;
	for (var i = 0; i < list.length; i++) {
		list[i] = list[i].replace(/\s/g, '');
		if (whiteRe.test(list[i])) { // XXX if these are duplicated by existing adblock.white entries, they'll be overwritten, when the whitelist is loaded (next) -- which is a good thing (just so we're clear).
			var alone = list[i].replace(whiteRe,''); // leave the original untouched, in case we don't import/remove it
			if (patRe.test(alone) || simpRe.test(alone) || !periodRe.test(alone)) continue;
			//url.spec = list[i];
			//whitelist[url.spec]=new Date().valueOf(), whitehosts[url.host]||(whitehosts[url.host]={}), whitehosts[url.host][url.spec]=true;
			whiteImported[alone]=true;
			list.splice(i--, 1); // remove imported-entry from list + back the counter (from splice)
		}
		else if (wpRe.test(list[i])) { // full whitelist filters -- these live in the pref-window's filterlist; leave 'em.
			alone = list[i].replace(wpRe,'');
			if (!patterns.white) patterns.white={};
			addPattern(alone, "", patterns.white);
		}
		else if (list[i].charAt(0)=="!") continue; // ignore comment-disabled filters
		else if (list[i].match(/^https?:\/\/[^*\/]+\//)) {
			url.spec = list[i];
			addPattern(list[i], url.host, patterns); }
		else
			addPattern(list[i], "", patterns);
	}
	// save list -- on truncation, or pref-change
	if (list.length < i || prefName) 
		try {
			Branch.setCharPref("patterns", list.join(" "));
			// XXX Disabled, so we don't accidentally lose our 'imported' whitelist filters, if we crash before the whitelist loads + saves (next)
			//prefObj.savePrefFile(null); // save the prefs to disk 
		} catch(e) {}
}

// Right-click > selection-import
function importFiltersFromString(input) {
	// now: unix + mac + dos environment-compatible
	var breakRe = /(?:\[Ad[Bb]lock\])((( +)|(\n+)|(\r+))+)/m; // updated, to handle space-delimiting (firefox1.5)
	var match = input.match(breakRe); // first: whole match -- second: backref-1 -- etc..
	//if (!match) input = input.replace(/ +/g,"\n"), match = input.match(breakRe);
	if (!match) { 
		if (!lastWindow||lastWindow.closed) lastWindow = windowMediator.getMostRecentWindow("navigator:browser");
		lastWindow.alert("The selected text could not be imported. \n\nMake sure you've selected the complete filterset.");
		return; }
	var linebreak = match[1];
	var inputArray = input.split(linebreak);
	
	var headerRe = /\[Ad[Bb]lock\]/; // tests if the first line is adblock's header
	if (inputArray.length > 1 && headerRe.test(inputArray[0])) {
		inputArray.shift();
		var emptyRe = /^\s*$/;
		if (emptyRe.test(inputArray.join(""))) var msg = "Make sure you've selected the complete filterset. Then, try again.";
		else {
			if (!lastWindow||lastWindow.closed) lastWindow = windowMediator.getMostRecentWindow("navigator:browser");
			var cancelled = !lastWindow.confirm("You're about to Import an Adblock filterset.\n\n..Do you want to continue?");
			if (cancelled) return; // first confirm -- user cancelled.
			//var msg = "Do you really want to import the selected filterset?", title = "Adblock Import", retainLabel = this.keepTxt, removeLabel = this.clearTxt;
						//var msg = "Do you want to Append to your current filters?\n..or Replace them?", title = "Adblock Import", appendLabel = "Append", replaceLabel = "Replace", cancelLabel="Cancel";
						//var flags = promptService.BUTTON_TITLE_IS_STRING * promptService.BUTTON_POS_0 + promptService.BUTTON_TITLE_IS_STRING * promptService.BUTTON_POS_1 + promptService.BUTTON_TITLE_CANCEL * promptService.BUTTON_POS_2 + promptService.BUTTON_POS_2_DEFAULT;
						//var buttonPressed = promptService.confirmEx(lastWindow,title,msg,flags,appendLabel,replaceLabel,cancelLabel,null,{});
			var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService); 
			var msg = "Do you want to Replace your current filters?\n..or Append to them?", title = "Adblock Import", appendLabel = "Append", replaceLabel = "Replace", cancelLabel="Cancel Import"; 
			var flags = promptService.BUTTON_TITLE_IS_STRING * promptService.BUTTON_POS_2 + promptService.BUTTON_TITLE_IS_STRING * promptService.BUTTON_POS_1 + promptService.BUTTON_TITLE_IS_STRING * promptService.BUTTON_POS_0 + (promptService.BUTTON_POS_0_DEFAULT?promptService.BUTTON_POS_0_DEFAULT:0); 
			if (!lastWindow||lastWindow.closed) lastWindow = windowMediator.getMostRecentWindow("navigator:browser");
			var buttonPressed = promptService.confirmEx(lastWindow,title,msg,flags,cancelLabel,replaceLabel,appendLabel,null,{});
			if (buttonPressed == 0) return; // second confirm -- user cancelled.
			var shouldAppend = (buttonPressed == 2);
			//var shouldAppend = !lastWindow.confirm("Do you want to overwrite the current list?\n..if not, pressing 'cancel' will append.");
			loadFilters("filter-Import", inputArray, shouldAppend);
			loadWhitelist("filter-Import"); // XXX ..i don't think there's any real reason to keep the 'prefName' argument, here.
			return; }
	}
	else msg = "Perhaps its line-returns aren't formatted correct- drop a line to the Adblock Forums, and we'll take a look.";
	if (!lastWindow||lastWindow.closed) lastWindow = windowMediator.getMostRecentWindow("navigator:browser");
	lastWindow.alert("The selected text could not be imported. \n\n"+msg);
}

var checkPrefsFlag = false;
function checkPrefs() {
	if (checkPrefsFlag) return;
	
	// get profile directory
	try{
		var dirService = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties);
		var file = dirService.get("ProfD", Components.interfaces.nsIFile);
		file.append("pref");
		file.append("adblock.js");  }
	catch(e){ return; }
	
	checkPrefsFlag = true; // root-install/backstagepass wont have access to the profile, until later -- duh.
	
	// if not Mozilla - def prefs work
	if (!file.exists())
		return;
	
	// read data
	var data = "";
	var fstream = Components.classes["@mozilla.org/network/file-input-stream;1"].createInstance(Components.interfaces.nsIFileInputStream);
	var sstream = Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(Components.interfaces.nsIScriptableInputStream);
	fstream.init(file, 1, 0, false);
	sstream.init(fstream); 
	
	var str = sstream.read(-1);
	while (str.length > 0) {
	  data += str;
	  str = sstream.read(-1); }
	
	sstream.close();
	fstream.close();
	
	try{ file.remove(true); }catch(e){}// only use the file once; after that, any update-install can re-instate it
	
	// do something with read data
	var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);	
	
	// remove comments and split lines into array
	data = data.replace(/\/\/.*/gm, "").replace(/\/\*([^*][^\/])*[^\0]?\*\//gm, "");
	data = data.split(';');
	
	for (var i=0; i<data.length; i++) {
		//if valid pref line
		if (/^\s*pref\(.*\,.*\)\s*$/i.test(data[i])) {
			var pref = data[i].slice(data[i].indexOf('(')+2, data[i].indexOf(',')-1);
			var value = data[i].slice(data[i].indexOf(',')+1, data[i].indexOf(')'));
			
			if (/^\s*false\s*$/i.test(value) || /^\s*true\s*$/i.test(value)) { // if boolean pref
				value = (/\s*true\s*/i.test(value));
				try{ prefs.getBoolPref(pref); }catch(e){ prefs.setBoolPref(pref, value); }
			}
			else if (/^\s*["][^"]*["]\s*$/.test(value)) { // if char pref
				value = value.slice(value.indexOf('"')+1,value.lastIndexOf('"'));
				try {prefs.getCharPref(pref);} catch (e) {prefs.setCharPref(pref, value);}
			}
			else { // if integer pref
				value = new Number(value);
				try {prefs.getIntPref(pref);} catch (e) {prefs.setIntPref(pref, value);}
			}
		}
	}
}

function addWhite(data,toggle) {
	var stuff     = data.split(" ");
	var url       = stuff[0]
	var whiteType = stuff[1];
	var prevType  = stuff[2];
	var stdURL = Components.classes[STDURL_CONTRACTID].createInstance(Components.interfaces.nsIURI);
		stdURL.spec = url;
	var host = stdURL.host;
		if (host) host = host.replace(/^www\./,"");
	var date = new Date().valueOf();
	
	if (url=="about:blank") return; // ..come to think of it, we should disable the menuitems for chrome/file sites, and such
	if (toggle && isWhite(url)==whiteType) whiteType = 'null'; // for the browser menu-items (adblock.xul)
	switch(whiteType) {
		case "null": // clear whitelisting -> host+page
			//var pages = whitehosts[host];
			//for (var n in pages) delete whitelist[n];
			//delete whitehosts[host];
			if (whitehosts[host]) delete whitehosts[host][url];
			delete whitelist[host];
			delete whitelist[url];       break;
		case "0":   // whitelist this page
			delete whitelist[host];
			whitehosts[host]||(whitehosts[host]={});
			whitehosts[host][url]=true;
			whitelist[url]=date;         break;
		case "1":   // whitelist this host
				//pages = whitehosts[host];
				//for (n in pages) delete whitelist[n];
				//delete whitehosts[host];
			if (whitehosts[host]) delete whitehosts[host][url];
			delete whitelist[url]; // only delete the local url -- other, individually whitelisted-pages will be retained.
			whitelist[host]=date; break;
	}
	
	saveWhitelist();
}

// Converts a pattern into RegExp and adds it to the list
function addPattern(pattern, host, patterns) {
	if (typeof(patterns[host]) == 'undefined')
		patterns[host] = [];
	if (typeof(patterns['all']) == 'undefined')
		patterns['all'] = [];

	var regexp;
	if (pattern.charAt(0) == "/" && pattern.charAt(pattern.length - 1) == "/")  // pattern is a regexp already
		regexp = pattern.substr(1, pattern.length - 2);
	else
		regexp = pattern.replace(/^[\*]*(.(?:[^\*]|[\*]+[^\*])*)[\*]*$/, "$1").replace(/[\*]+/, "*").replace(/([^\w\*])/g, "\\$1").replace(/\*/g, ".*") + "(?:\\n)?"; // append flag for "adblock simple-filter" -- for filterall's de-regexp
		//regexp = pattern.replace(/(?:^\*).|.(?:\*$)/g, "").replace(/([^\w\*])/g, "\\$1").replace(/\*/g, ".*") + "(?:\\n)?"; // append flag for "adblock simple-filter" -- for filterall's de-regexp
		//regexp = "^" + pattern.replace(/([^\w\*])/g, "\\$1").replace(/\*/g, ".*") + "(ABsf)?" + "$";

	try {
		regexp = new RegExp(regexp, "i");
		
		patterns[host].push(regexp);
		patterns['all'].push(regexp);
	} catch(e) {}
}


/*
 * Object frames
 */

// Creates a frame around a new object node
function makeObjectFrame(node, contentURL, wnd) {
	
	if (node.hasAttribute("AdblockFramedObject")) return;
	
	// Create frame node
	var frame, subDiv1, subDiv2, span;
	frame = node.ownerDocument.createElement('div'); 
	subDiv1 = node.ownerDocument.createElement('div');
	subDiv2 = node.ownerDocument.createElement('div');
	span = node.ownerDocument.createElement('span');
	
	// Style frame node
	frame.setAttribute("style", style="margin: 0px; padding: 0px; overflow: visible;");
	subDiv1.setAttribute("style", "height: 0px; width: 100%; overflow: visible;");
	subDiv2.setAttribute("style", "padding: 1px; vertical-align: bottom; border-style: ridge ridge none ridge; border-width: 2px 2px 0px 2px; -moz-border-radius-topleft: 10px; -moz-border-radius-topright: 10px; -moz-opacity: 0.5; background-color: white;  position: relative; top: -19px; left: -5px; z-index: 900;  width: 48px; height: 15px;  cursor: pointer; overflow: visible;");
	span.setAttribute("style", "font-family: Arial,Helvetica,Sans-serif; font-size: 12px; font-style: normal; font-variant: normal; font-weight: normal; line-height: 140%; text-align: right; text-decoration: none; -moz-opacity: 1.5; color: black;");
	subDiv1.setAttribute("align", "right"); // right-align the tab.
	subDiv2.setAttribute("align", "center"); // center tab-text.
	
	// Adjust frame-width
	frame.style.width = node.getAttribute('width'); 
	frame.style.display = 'block';
	
	// Assemble frame node
	frame.appendChild(subDiv1);
	subDiv1.appendChild(subDiv2);
	subDiv2.appendChild(span);
	span.appendChild(node.ownerDocument.createTextNode('Adblock'));
	
	// Prevent Object Frame being added multiple times.
	node.setAttribute("AdblockFramedObject", true);
	
	// Add cross-node references
	var callback = function(){onBlockObject(node, contentURL, wnd);};
	var getFrame = function(){return frame;};
	var getNode  = function(){return node;};
	frame._AdblockObject = node; // add Node reference to frame
	if (node.nodeName.toLowerCase() != "applet") {
		node._AdblockFrame = frame; // objects: add Frame reference to node
		node.setAttribute("AdblockFramedObject2", true); }
	//else
		//node.setAttribute("AdblockFramedObject", true);
	frame.setAttribute("AdblockFrame", true); // to identify the outer-frame on tab-click
	subDiv2.addEventListener('click', callback, false); // event listener, for tab-click
	var frameName = "adblock-frame-n"+frameCounter++;
	frame.id = frameName;
	node.setAttribute("AdblockFrameName", frameName); // links applets to their frame, via unique-id
	
	var style = node.ownerDocument.defaultView.getComputedStyle(node, "");
	if (!frame.style.width) frame.style.width = style.getPropertyValue("width");
	
	var currentOffset = node, y = node.offsetTop, evalCode;
	
	while (currentOffset.offsetParent) {
		y += currentOffset.offsetParent.offsetTop;
		currentOffset = currentOffset.offsetParent; }
	
	if (y < 20) {
		var subFrame = frame.firstChild, subFrame2 = frame.firstChild.firstChild;
		subFrame2.style.setProperty("-moz-border-radius-bottomleft", "10px", null);
		subFrame2.style.setProperty("-moz-border-radius-bottomright", "10px", null);
		subFrame2.style.setProperty("border-style", "none ridge ridge ridge", null);
		subFrame2.style.setProperty("border-width", "0px 2px 2px 2px", null);
		subFrame2.style.setProperty("top", "0px", null);
		subFrame2.style.setProperty("right", "-5px", null);
		subFrame2.style.removeProperty("left");
		subFrame2.style.removeProperty("-moz-border-radius-topleft");
		subFrame2.style.removeProperty("-moz-border-radius-topright");
		subFrame.setAttribute("align", "left");
		
		if (node.nextSibling)
			  evalCode = "node.parentNode.insertBefore(frame, node.nextSibling);";
		else evalCode = "node.parentNode.appendChild(frame);";
	}
	else evalCode = "node.parentNode.insertBefore(frame, node);";
	
	var parentiFrame = node.ownerDocument.defaultView.frameElement;
	if (parentiFrame != null) {
		var nodeHeight = parseInt(style.getPropertyValue("height").replace(/(\d+).*$/, "$1"));
		var parentiFrameStyle = node.ownerDocument.defaultView.getComputedStyle(parentiFrame, "");
		var parentiFrameHeight = parseInt(parentiFrameStyle.getPropertyValue("height").replace(/(\d+).*$/, "$1"));
		
		if (nodeHeight + 20 + y > parentiFrameHeight) {
			parentiFrame.style.setProperty("height", nodeHeight+20+y, "important");
			parentiFrame.height = nodeHeight+20+y;
		}
	}
	
	eval(evalCode);
	node.style.removeProperty("visibility");
}

// Event handler, Adblock Tab has been clicked
function onBlockObject(node, contentURL, wnd) {
	/*node.ownerDocument.defaultView.top*/
	windowMediator.getMostRecentWindow("navigator:browser").openDialog("chrome://adblock/content/addfilterdialog.xul","Add filter", "chrome,modal,centerscreen", contentURL, node, wnd);
}
