/*
** GPL Licensed, by: Henrik, rue
*/

const DATASOURCE_CONTRACTID = '@mozilla.org/rdf/datasource;1?name=in-memory-datasource';
const RDFSERVICE_CONTRACTID = '@mozilla.org/rdf/rdf-service;1';
const CONTAINERUTILS_CONTRACTID = '@mozilla.org/rdf/container-utils;1';
const CONTENTPOLICY_CONTRACTID = "@mozilla.org/layout/content-policy;1";

var rdf =  Components.classes[RDFSERVICE_CONTRACTID].getService(Components.interfaces.nsIRDFService);
var utils = Components.classes[CONTAINERUTILS_CONTRACTID].getService(Components.interfaces.nsIRDFContainerUtils);
var policy = Components.classes[CONTENTPOLICY_CONTRACTID].getService(Components.interfaces.nsIContentPolicy);
var datasource, sequence;

const prefix = 'http://adblock.mozilla.org/rdf#';

var console = Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService);
function log(msg) { console.logStringMessage(msg); }

var nodes = {};
var nodeTypes = {};

var appShell  = Components.classes["@mozilla.org/appshell/appShellService;1"].getService(Components.interfaces.nsIAppShellService);
var hiddenWnd = appShell.hiddenDOMWindow; // global hidden-window
var apiConstants     = hiddenWnd.apiConstants;
var isBlocked        = hiddenWnd.isBlocked;
var isWhite          = hiddenWnd.isWhite;
var addWhite         = hiddenWnd.addWhite;
var storeAdblockData = hiddenWnd.storeAdblockData;
var whiteKey         = hiddenWnd.whiteKey;
var treeConstants = new Object(); // the tree-object changed, per bug 221619 -- this will hold our "api-switch" bool

var typeDescr = [];
typeDescr[apiConstants.SCRIPT]          = 'SCRIPT';
typeDescr[apiConstants.IMAGE]           = 'IMAGE';
typeDescr[apiConstants.SUBDOCUMENT]     = 'SUBDOCUMENT';
typeDescr[apiConstants.OBJECT]          = 'OBJECT';
typeDescr[apiConstants.DOCUMENT]        = 'DOCUMENT';
typeDescr[apiConstants.XMLHTTPREQUEST]  = 'AJAX';

var flasher = {
	url: null,
	loopCount: 0,
	minorCount: 0,
	hiddenCount: 0,
	timeout: null,
	origOutline: [],

	flash: function(url) {
		this.stop();
		this.url = url;
		this.loopCount = 0;
		this.minorCount = 0;
		this.hiddenCount = 0;
		this.doFlash();
	},

	doFlash: function() {
		// if we've looped the max-times, stop
		if ( ! (this.loopCount < 3)) {
			this.switchOff();
			this.url = null;
			return; }

		// If we're still visible-flashing
		if (this.minorCount < 12) {
			if (this.minorCount % 2) this.switchOff();
			else this.switchOn();
			this.minorCount++; }
		// ..or we've flashed our visible-limit
		else {
			// ..and we're still hidden-flashing
			if (this.hiddenCount < 6)
				this.hiddenCount++;
			// ..otherwise, we've flashed our hidden-limit
			else {
				this.loopCount++;
				this.minorCount = 0; 
				this.hiddenCount = 0; }
		}
		
		this.timeout = setTimeout('flasher.doFlash()', 120);
	},

	stop: function() {
		if (this.url != null) {
			clearTimeout(this.timeout);
			this.switchOff(); }
	},

	switchOn: function() {
		var list = nodes[this.url];
		for (var i=0; i<list.length; i++) {
			var node = list[i];
			if (typeof(node.contentDocument) != 'undefined' && node.contentDocument 
					&& typeof(node.contentDocument.getElementsByTagName) == 'function' 
					&& node.contentDocument.getElementsByTagName('body') 
					&& node.contentDocument.getElementsByTagName('body').length > 0
				)
				node = node.contentDocument.getElementsByTagName('body')[0];
			//this.origOutline[String(i)] = node.style.MozOutline;
			if (node.style) node.style.MozOutline = "#AB0000 dotted 2px !important";
		}
	},

	switchOff: function() {
		var list = nodes[this.url];
		for (var i=0; i<list.length; i++) {
			var node = list[i];
			if (typeof(node.contentDocument) != 'undefined' && node.contentDocument 
					&& typeof(node.contentDocument.getElementsByTagName) == 'function' 
					&& node.contentDocument.getElementsByTagName('body') 
					&& node.contentDocument.getElementsByTagName('body').length > 0
				)
				node = node.contentDocument.getElementsByTagName('body')[0];
			//node.style.MozOutline = this.origOutline[String(i)] ? this.origOutline[String(i)] : 'none';
			if (node.style) node.style.MozOutline = 'none';
		}
	}
};

// if the user clicks "OK", saves the filter and refilters the document
function allOK() {
	
	var filterAccepted = tryToAddFilter();
	
	if(!filterAccepted)
		return false; // don't stop filter if the user cancels the addition
	
	updateWhitelist();
	flasher.stop();
	return true;
}

var url, initialWhiteType=null;
function setupWhitelist(wnd) {
	url             = wnd.location+"";
	var whiteType   = initialWhiteType = isWhite(url);
	var whiteActive = (whiteType==null) ? false:true;
	if (whiteActive) document.getElementById("whitelist-menu").selectedIndex = whiteType;
	document.getElementById("whitelist-menu").removeAttribute('sizetopopup');   // since moz1.3.1 overrides our xul on this -
	document.getElementById("whitelist-checkbox").setAttribute("checked", whiteActive);
		var c = document.getElementById("whitelist-container");
		/*if (!c.hasAttribute('style'))*/ {
			var cHeight = window.getComputedStyle(c, null).getPropertyValue('height');
			c.setAttribute('style', "min-height: "+cHeight+" !important;"); }
	document.getElementById("whitelist-checkbox").doCommand();
	if (wnd.document[whiteKey]) document.getElementById("adblock-treechildren").setAttribute('class','whitelistedNonMutable')
}
function updateWhitelist() {
	var whiteType   = currentWhiteType(); // our current setting
	var stuff       = [url,whiteType].join(" ");
	addWhite(stuff);
}
function currentWhiteType() {
	var checkValue  = document.getElementById("whitelist-checkbox").getAttribute("checked");
	var whiteActive = (checkValue == "true") ? true:false;
	var whiteType   = whiteActive ? document.getElementById("whitelist-menu").value : "null";
	return whiteType;
}

function whiteCheckboxCommand(item) {
	var checked = (item.getAttribute('checked')=='true');
	var p = item.parentNode.parentNode; 
	p.setAttribute('class',checked?'nontranslucent':'translucent');
	
	var m = document.getElementById('whitelist-menu');
	m.disabled = !checked;
	m.setAttribute('style',(checked?' ':'display:none')); 
	//m.setAttribute('style',(checked?'visibility:visible':'visibility:hidden;width:0;max-width:0;padding-left:0;padding-right:0;margin-left:0;margin-right:0;border-left-style:none;border-right-style:none;')); 
	
	if (checked) { 
		var mp = m.parentNode; mp.removeChild(m), mp.appendChild(m); } 
	if (m.hasAttribute('sizetopopup')) m.removeAttribute('sizetopopup');  
	
	var r = document.getElementById('root-vbox'); 
	checked ? r.setAttribute('whitelisted',true) : r.removeAttribute('whitelisted');  
	
	var t = document.getElementById('adblock-treechildren'); 
	if (!t.hasAttribute('class')||t.getAttribute('class')!='whitelistedNonMutable') checked ? t.setAttribute('class','whitelisted') : t.removeAttribute('class');   
	document.getElementById('references-tree').treeBoxObject.clearStyleAndImageCaches();   
	
	var textbox = document.getElementById("filter");
	var tree = document.getElementById('references-tree');
	treeConstants.oldStyleAPI ? tree.treeBoxObject.selection.clearSelection() : tree.treeBoxObject.view.selection.clearSelection();
	textbox.value = "";
	textbox.disabled = false;
}

function whiteMenuClick(item) {
	var c = document.getElementById('whitelist-checkbox'); 
	if (c.getAttribute('checked') != 'true') 
		c.setAttribute('checked', true),
		c.doCommand();   /*, item.open=true;*/
}

function tryToAddFilter() {
	var textbox = document.getElementById("filter");
	
	if (!textbox.disabled) { // only if an already-filtered item wasn't selected
		var pattern = textbox.value;
		if (pattern.charAt(0) == "/" && pattern.charAt(pattern.length - 1) == "/" && !adblockWarnRegExp())
			return false; // if the user didn't want to enter a regular expression, cancel the "OK"
			
		var emptyRe = /^\s*$/;
		if (pattern == null || emptyRe.test(pattern)) return true; // don't add null filter
		
		addFilter(pattern); // save the filter to prefs
		var wnd = window.arguments[0];
		refilter(wnd); // refilter the whole page -- we'll handle wnd.closed conditions during refiltering
	}
	
	return true;
}

function allCancel() {
	flasher.stop();
	return true;
}

// fill the list with external references
function init() {
	datasource = Components.classes[DATASOURCE_CONTRACTID].createInstance(Components.interfaces.nsIRDFDataSource);
	
	var tree = document.getElementById('references-tree');
	tree.database.AddDataSource(datasource);

	sequence = utils.MakeSeq(datasource, rdf.GetResource('urn:root:references'));

	/*
	if (window.arguments[1])
		addAllReferences(window.arguments[0].ownerDocument.defaultView.top);
	else
		addReferences(window.arguments[0]);
	*/
	
	var wnd = window.arguments[0];
	
	setupWhitelist(wnd);
	addAllReferences(wnd); // get all items for current browser-page
	
	treeConstants.oldStyleAPI = (tree.treeBoxObject.selection ? true : false); // per bug 221619 (see above)
	treeConstants.oldStyleAPI ?
		tree.treeBoxObject.selection.clearSelection() //.select(null); //0)
		:
		tree.treeBoxObject.view.selection.clearSelection(); //.select(null); //0)
	if ('focus' in tree) tree.focus();
	if ('clearStyleAndImageCaches' in tree.treeBoxObject) 
		tree.treeBoxObject.clearStyleAndImageCaches(); // update treechildren styles, in case whitelisting was set (or odd column-alignment adjustment, under moz1.3.1 -- persistence?)
}

// on selection-change, show the item's url (if unfiltered), OR the filter that caught the item
function onSelectionChange() {
	var textbox = document.getElementById('filter');
	var tree = document.getElementById('references-tree');
	var selection = treeConstants.oldStyleAPI ?
		tree.treeBoxObject.selection
		:
		tree.treeBoxObject.view.selection;
	var editButton = document.getElementById("edit-filter");

	if (selection.count != 1 || !selection.isSelected(selection.currentIndex))
		return (editButton.hidden=true);
	
	//Exception is thrown when called on about:blank page.
	try {var resource = tree.builderView.getResourceAtIndex(selection.currentIndex);} catch(e) {return;}
	
	var filter = datasource.GetTarget(resource, rdf.GetResource(prefix + 'filter'), true);
	if (filter) filter = filter.QueryInterface(Components.interfaces.nsIRDFLiteral);
	if (filter && filter.Value != "NoMatch") {
		filter = filter.Value.toString();
		var simpleRe = /\(\?\:\\n\)\?/; // test for the "Adblock simple-filter" flag
		//var simpleRe = /(^\^http)|(\(ABsf\)\?\$\/\i$)/;
		// if the filter is simple, de-RegExp -- otherwise, get the pattern sans-delimiters
		if (simpleRe.test(filter)) filter = filter.replace(/^(\@\@)?\/(.*)\(\?\:\\n\)\?\/i?$/, "$1$2").replace(/\\\./g, " ").replace(/\.\*/g, "*").replace(/ /g, ".").replace(/\\/g, ""); // first 'replace' is same as substr, below -- left regexp for fun
		else filter = filter.substr(0, filter.length - 1); // length isn't zero-based + / + $ + flag "i" are anchored to end
		if (!textbox.disabled) textbox.disabled = true;
		textbox.value = filter;
		editButton.hidden=false;
		flasher.stop(); }
	else {
		var url = datasource.GetTarget(resource, rdf.GetResource(prefix + 'url'), true);
		if (url) url = url.QueryInterface(Components.interfaces.nsIRDFLiteral);
		if (url) {
			if (textbox.disabled) textbox.disabled = false;
			textbox.value = url.Value;
			editButton.hidden=true;
			flasher.flash(url.Value); }
	}
}

function addNodeReferences(data) {

	var url = data.contentLocation.spec;
	var filterMatch = data.filterMatch;
	var node = data.node;
	var type = data.contentType;
	
	if (!filterMatch) {
		if (initialWhiteType==null) var displayStyle = "unfiltered";
		else displayStyle = "white";  }
	else if (filterMatch.white) displayStyle = "white", filterMatch = filterMatch.asWhiteString;
	else displayStyle = "filtered"; 

	if (typeof(nodes[url]) == 'undefined') {
		nodes[url] = [];
		nodeTypes[url] = [];
		
		var resource = rdf.GetResource(url);
		datasource.Assert(resource, rdf.GetResource(prefix + 'url'), rdf.GetLiteral(url), true);
		datasource.Assert(resource, rdf.GetResource(prefix + 'type'), rdf.GetLiteral(typeDescr[type].toLowerCase()), true);
		datasource.Assert(resource, rdf.GetResource(prefix + 'elements'), rdf.GetLiteral(''), true);
		datasource.Assert(resource, rdf.GetResource(prefix + 'displaystyle'), rdf.GetLiteral(displayStyle), true);
		datasource.Assert(resource, rdf.GetResource(prefix + 'filter'), rdf.GetLiteral(filterMatch ? filterMatch : "NoMatch"), true);
		
		sequence.AppendElement(resource);
	}
	

	for (var i=0; i<node.length; i++) {
		try{ var name = node[i].tagName; }catch(e){name=null;}
		
		//Handle background images separately since the object has no tag name
		try{ if (!name && type == apiConstants.IMAGE && node[i].toString().indexOf("HTMLDocument")!=-1)
				name = "BGND";  }catch(e){ log('addNodeReferences  ->  node: '+node[i]+' , url: '+url); continue; }
		
		var found = false;
		var types = nodeTypes[url];
		for (var j=0; j<types.length; j++) {
			if (types[j] == name) {
				found = true;
				break; }
		}
		
		if (!found) {
			var prev = types.join(', '); // joins the types-array with comma separation
			types.push(name);
			
			datasource.Change(rdf.GetResource(url), rdf.GetResource(prefix + 'elements'), rdf.GetLiteral(prev), rdf.GetLiteral(types.join(', ')));
		}
		nodes[url].push(node[i]);
	}
}

// fill the list with all external references in a window
function addAllReferences(wnd) {
	var data = /*wnd._Adblock||*/wnd.document._Adblock; //||{};
		//log("filterall->\nwnd: "+wnd+"\ndocument: "+wnd.document+"\n _Adblock: "+wnd.document._Adblock);
	if (!data) eval("data=wnd.document._Adblock;");
	if (!data) return;
	for (var i in data)
		addNodeReferences(data[i]);
}

// adds the filter to prefs
function addFilter(filter) {
	if (filter != "") {
		var wPatterns = null;
		try {
			var prefObj = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
			var Branch = prefObj.getBranch("adblock.");
			wPatterns = Branch.prefHasUserValue("patterns") ? Branch.getCharPref("patterns") : null;
		} catch(e) {}
		
		if (wPatterns) wPatterns += " " + filter;
		else wPatterns = filter;
		try {
			Branch.setCharPref("patterns", wPatterns);
			//prefObj.savePrefFile(null); // save the prefs to disk -- component.js "loadSettings()" handles this now
			var observerService = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
			observerService.notifyObservers(null, "Adblock-PrefChange", "FilterChange");
			observerService.notifyObservers(null, "Adblock-SavePrefFile", null);
		} catch(e) {}
	}
}

// refilters all nodes of a window
function refilter(wnd) {
	var wasActive = opener.toggleObjectOverride(wnd,false);
	var doc = wnd.document;
	var data = /*wnd._Adblock||*/doc._Adblock||{};
	try{ doc[whiteKey] = (currentWhiteType()!='null') ? true:false; }catch(e){}
	//if (doc._Adblock) doc._Adblock=null; // XXX since the ordering in the obj is maintained, we can actually leave this
	// we flush the data below -- delete data[i]
	var i, node, contentType, contentLocation;
	
	for (i in data) {
		if (wnd.closed) return; // note: this is *inside the loop -- the window might be closed, mid-run
		node = data[i].node;
		contentType = data[i].contentType;
		contentLocation = data[i].contentLocation;
		delete data[i];
		if (contentType == apiConstants.XMLHTTPREQUEST) { 
			for (var k=0; k<node.length; k++)
				storeAdblockData(node[k], contentLocation, contentType, isBlocked(contentLocation), null,doc); 
			continue; // we don't reset ajax-requests; they can't be 'refiltered', and there's no easy way to re-create the node
		}
		
		// the "immediate" boolean directs non-delayed collapsation
		for (var j=0; j<node.length; j++)
			if (node[j] && node[j].tagName && node[j].tagName=='XMLHttpRequest') storeAdblockData(node[k], contentLocation, contentType, isBlocked(contentLocation), null,doc); // some latent ajax-requests are grouped with non-ajax requests for the same content-location -- so they wont have been caught, above.
			else policy.shouldLoad(contentType, contentLocation, apiConstants.oldStyleAPI?node[j]:null, apiConstants.oldStyleAPI?wnd:node[j], apiConstants.oldStyleAPI?((node._AdblockImmediate=true)?null:null):apiConstants.IMMEDIATE, null);
	}
	var callback = function(chrome, content){ 
		if (!chrome.closed && !content.closed && content.document) 
			chrome.toggleObjectOverride(content,true); 
	};
	if (wasActive) opener.setTimeout(callback, 500, opener,wnd);
}

